{ Unit: BcAPITimerObject
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    TObject version of the TBcAPITimer.

  History:

    2001-11-27: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcAPITimerObject;

interface

uses Windows, Classes, SysUtils;

const
  DefaultEnabled = False; // this is for the object class as it's created dynamically always!
  DefaultInterval = 1000;

type
  TBcTimerEvent = procedure(Sender: TObject) of object;

  TBcAPITimerObject = class(TObject)
  private
    FEnabled: Boolean;
    FInterval: Cardinal;
    FOnTimer: TBcTimerEvent;
    FTimerHandle: THandle;
    FSender: TObject;
    procedure SetEnabled(const Value: Boolean);
    procedure SetInterval(const Value: Cardinal);
    procedure BeginTimer;
    procedure EndTimer;
  public
    constructor Create(ASender: TObject); virtual;
    destructor Destroy; override;

    class procedure PackRef;
    class procedure IncRef(ATimerHandle: THandle; ATimerObject: TBcAPITimerObject);
    class procedure DecRef(ATimerHandle: THandle);
    class function FindTimerObject(ATimerHandle: THandle): TBcAPITimerObject;

    property TimerHandle: THandle read FTimerHandle;
    property Enabled: Boolean read FEnabled write SetEnabled default DefaultEnabled;
    property Interval: Cardinal read FInterval write SetInterval default DefaultInterval;

    property OnTimer: TBcTimerEvent read FOnTimer write FOnTimer;
  end;

implementation

{ TBcAPITimerObject }

type
  PTimerRef = ^TTimerRef;
  TTimerRef = record
    TimerHandle: THandle;
    TimerObject: TBcAPITimerObject;
  end;

  TTimerRefTable = class(TList)
  public
    destructor Destroy; override;
  end;

var
  TimerRefTable: TTimerRefTable;

procedure APITimerCallback(hWnd: THandle; uMsg: Cardinal; TimerId: THandle;
  dwTime: Cardinal); stdcall;
var Sender: TBcAPITimerObject;
begin
  Sender := TBcAPITimerObject.FindTimerObject(TimerId);
  if Assigned(Sender) and Assigned(Sender.FOnTimer) then
    Sender.FOnTimer(Sender.FSender);
end;

procedure TBcAPITimerObject.BeginTimer;
begin
  if FTimerHandle <> 0 then
    EndTimer;
  FTimerHandle := SetTimer(0, 0, FInterval, @APITimerCallback);
  IncRef(FTimerHandle, Self);
end;

constructor TBcAPITimerObject.Create(ASender: TObject);
begin
  FEnabled := DefaultEnabled;
  FInterval := DefaultInterval;
  FSender := ASender;
  if not Assigned(FSender) then
    FSender := Self;
end;

class procedure TBcAPITimerObject.DecRef(ATimerHandle: THandle);
var I: Integer;
begin
  for I := 0 to TimerRefTable.Count - 1 do
  begin
    if  PTimerRef(TimerRefTable[I])^.TimerHandle = ATimerHandle then
    begin
      FreeMem(TimerRefTable[I]);
      TimerRefTable.Delete(I);
      TimerRefTable.Capacity := TimerRefTable.Count;
      Break;
    end;
  end;
end;

destructor TBcAPITimerObject.Destroy;
begin
  EndTimer;
  inherited Destroy;
end;

procedure TBcAPITimerObject.EndTimer;
begin
  if FTimerHandle <> 0 then
  begin
    KillTimer(0, FTimerHandle);
    DecRef(FTimerHandle);
    FTimerHandle := 0;
  end;
end;

class function TBcAPITimerObject.FindTimerObject(
  ATimerHandle: THandle): TBcAPITimerObject;
var I: Integer;
begin
  Result := nil;
  for I := 0 to TimerRefTable.Count - 1 do
  begin
    if PTimerRef(TimerRefTable[I])^.TimerHandle = ATimerHandle then
    begin
      Result := PTimerRef(TimerRefTable[I])^.TimerObject;
      Break;
    end;
  end;
end;

class procedure TBcAPITimerObject.IncRef(ATimerHandle: THandle;
  ATimerObject: TBcAPITimerObject);
var
  Ref: PTimerRef;
begin
//  PackRef;
  GetMem(Ref, SizeOf(TTimerRef));
  Ref.TimerHandle := ATimerHandle;
  Ref.TimerObject := ATimerObject;
  TimerRefTable.Add(Ref); // TTimerRefTable owns the Ref and frees it
end;

class procedure TBcAPITimerObject.PackRef;
begin
  TimerRefTable.Pack;
end;

procedure TBcAPITimerObject.SetEnabled(const Value: Boolean);
begin
  if FEnabled <> Value then
  begin
    FEnabled := Value;
    if FEnabled then
      BeginTimer
    else
      EndTimer;
  end;
end;

procedure TBcAPITimerObject.SetInterval(const Value: Cardinal);
begin
  if FInterval <> Value then
  begin
    FInterval := Value;
    if FEnabled then
      BeginTimer
    else
      EndTimer;
  end;
end;

{ TTimerRefTable }

destructor TTimerRefTable.Destroy;
var I: Integer;
begin
  for I := 0 to Count - 1 do
    FreeMem(Items[I]);
  inherited Destroy;
end;

initialization
  TimerRefTable := TTimerRefTable.Create;

finalization
  FreeAndNil(TimerRefTable);

end.
