{ Unit: BcStringsContainer
  ===========================================================================

  The contents of this file are subject to the Bluecave Limited License V 1.0
  (the "License"). You may not copy or use this file, in either source code
  or executable form, except in compliance with the License. You may obtain
  a copy of the License at http://www.bluecave.net/licenses/.

  Software distributed under the License is distributed on an "AS IS" basis,
  WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
  for the specific language governing rights and limitations under the
  License.

  Copyright (C) 2001-2002 Bluecave Software. All Rights Reserved.
  http://www.bluecave.net/

  Copyright (C) 2001-2002 Jouni Airaksinen. All Rights Reserved.
  http://Mintus.Codefield.com/ -- Mintus@Codefield.com

  ===========================================================================

  Unit description:

    Component to store strings easily in TStringLists.

    I was missing a design-time TStringList container so I made one and
    decided to make it properly in the first place. The container can
    contain multiple stringlists with custom naming.

    Should work with Delphi 6 too without changes but I'm not sure.


  Notes:

    For now Delimiters property works only with ValuesByIndex[] array
    property. You may define multiple one char Delimiters. First char found
    from the string is used. ValuesByIndex[] is good for random access.

    The Text and CommaText properties are read-only unlike the TStrings
    counterparts.

    There are properties to access the stringlists in TBcStringsContainer
    like they would be a one big stringlist. Note that changing the order
    (e.g. by Insert or Delete) of the stringlists naturally changes the
    indexing.

    There is a default action through component editor and you can save
    all the strings to text file through the context menu of the component.

  History:

    2001-10-01: Initial version

  =========================================================================== }

{$I DFS.inc}
{$I BcDirectives.inc}

unit BcStringsContainer;

interface

uses
  Windows, Messages, SysUtils, Classes, Controls;

const
  DefaultDelimiters = '=';
  DefaultUpdateDuplicates = False;
  DefaultRaiseErrors = False;

type
  TOnRefreshDuplicatesEvent = procedure(Sender: TObject; ADuplicates: TDuplicates; var AllowRefresh: Boolean) of object;

  TBcStringsContainer = class;

  TStringsItem = class(TCollectionItem)
  private
    FStrings: TStrings;
    FName: string;
    FUpdateDuplicates: Boolean;
    FOnRefreshDuplicates: TOnRefreshDuplicatesEvent;
    procedure SetStrings(const Value: TStrings);
    procedure SetName(const Value: string);
    function GetSorted: Boolean;
    procedure SetSorted(const Value: Boolean);
    function GetDuplicates: TDuplicates;
    procedure SetDuplicates(const Value: TDuplicates);
    procedure SetUpdateDuplicates(const Value: Boolean);
  protected
    function GetDisplayName: string; override;
  public
    constructor Create(Collection: TCollection); override;
    destructor Destroy; override;
    procedure Assign(Source: TPersistent); override;
    function GetNamePath: string; override;
    procedure RefreshDuplicates;
  published
    property Strings: TStrings read FStrings write SetStrings;
    property Name: string read FName write SetName;
    property Sorted: Boolean read GetSorted write SetSorted stored False;
    property UpdateDuplicates: Boolean read FUpdateDuplicates write SetUpdateDuplicates default DefaultUpdateDuplicates;
    property Duplicates: TDuplicates read GetDuplicates write SetDuplicates stored False;
    property OnRefreshDuplicates: TOnRefreshDuplicatesEvent read FOnRefreshDuplicates write FOnRefreshDuplicates;
  end;

  TStringsItems = class(TCollection)
  private
    FStringsContainer: TBcStringsContainer;
    function GetItem(Index: Integer): TStringsItem;
    procedure SetItem(Index: Integer; Value: TStringsItem);
  protected
    function GetOwner: TPersistent; override;
    procedure SetItemName(Item: TCollectionItem); override;    procedure Update(Item: TCollectionItem); override;

  public
    constructor Create(AStringsContainer: TBcStringsContainer);
    function Add: TStringsItem;
{$WARNINGS OFF} // Delphi 6 warns about this..
    property Owner: TBcStringsContainer read FStringsContainer;
{$WARNINGS ON}
    property Items[Index: Integer]: TStringsItem read GetItem write SetItem; default;
  end;

  TBcStringsContainer = class(TComponent)
  private
    { Private declarations }
    FItems: TStringsItems;
    FDelimiters: string;
    FRaiseErrors: Boolean;
    procedure SetItems(Value: TStringsItems);
    function GetItemsByName(Name: string): TStringsItem;
    function GetStringsCount: Integer;
    function GetStrings(Index: Integer): string;
    function GetValues(Name: string): string;
    function GetNames(Index: Integer): string;
    function GetValuesByIndex(Index: Integer): string;
    function GetCount: Integer;
    function GetText: string;
    function GetCommaText: string;
    procedure SetValues(Name: string; const Value: string);
  protected
    { Protected declarations }
    function ConvertToItemIndex(const Index: Integer;
      var ItemIndex, StringsIndex: Integer): Integer;
  public
    { Public declarations }
    constructor Create(AOwner: TComponent); override;
    destructor Destroy; override;

    procedure SaveToFile(const FileName: string);
    procedure Clear;

    property ItemsByName[Name: string]: TStringsItem read GetItemsByName; default;

    property Strings[Index: Integer]: string read GetStrings;
    property Values[Name: string]: string read GetValues write SetValues;
    property ValuesByIndex[Index: Integer]: string read GetValuesByIndex;
    property Names[Index: Integer]: string read GetNames;
    property StringsCount: Integer read GetStringsCount;
    property Count: Integer read GetCount;
    property Text: string read GetText;
    property CommaText: string read GetCommaText;
  published
    { Published declarations }
    property Items: TStringsItems read FItems write SetItems;
    property Delimiters: string read FDelimiters write FDelimiters;
    property RaiseErrors: Boolean read FRaiseErrors write FRaiseErrors default DefaultRaiseErrors;
  end;

implementation

{ TStringsItem }

procedure TStringsItem.Assign(Source: TPersistent);
begin
  if Source is TStringsItem then
  begin
    FStrings.Assign(TStringsItem(Source).FStrings);
    FName := TStringsItem(Source).FName;
  end
  else inherited Assign(Source);
end;

constructor TStringsItem.Create(Collection: TCollection);
begin
  inherited Create(Collection);
  FUpdateDuplicates := DefaultUpdateDuplicates;
  FStrings := TStringList.Create;
end;

destructor TStringsItem.Destroy;
begin
  FreeAndNil(FStrings);
  inherited Destroy;
end;

function TStringsItem.GetDisplayName: string;
var S: string;
begin
  if Assigned(FStrings) then
  begin
    S := FName;
    if S = '' then S := ClassName;
    Result := Format('%s (Lines: %d)', [S, FStrings.Count]);
  end else
    Result := inherited GetDisplayName;
end;

function TStringsItem.GetDuplicates: TDuplicates;
begin
  Result := TStringList(Strings).Duplicates;
end;

function TStringsItem.GetNamePath: string;
begin
{  if FName <> '' then
    Result := FName
  else}
    Result := inherited GetNamePath;
end;

function TStringsItem.GetSorted: Boolean;
begin
  Result := TStringList(Strings).Sorted;
end;

{ Note that this procedure will change the Strings pointer! So, if you have
  references, you should change them. Use the change event. }
procedure TStringsItem.RefreshDuplicates;
var
  TmpStrings: TStringList;
  Allow: Boolean;
  OldSorted: Boolean;
begin
  try
    { create new and copy properties }
    TmpStrings := TStringList.Create;

    Allow := True;
    if Assigned(OnRefreshDuplicates) then
      OnRefreshDuplicates(Self, TStringList(Strings).Duplicates, Allow);

    if Allow then
    begin
      TmpStrings.AddStrings(Strings);
      TmpStrings.OnChange := TStringList(Strings).OnChange;
      TmpStrings.OnChanging := TStringList(Strings).OnChanging;

      TStringList(Strings).OnChange := nil;
      TStringList(Strings).OnChanging := nil;

      TStringList(Strings).OnChange := TmpStrings.OnChange;
      TStringList(Strings).OnChanging := TmpStrings.OnChanging;

      with TStringList(Strings) do
      begin
        Clear;
        OldSorted := Sorted;
        Sorted := True;                      // duplicate update requires Sorted
        AddStrings(TmpStrings);              // Copy and sort/dup
        Sorted := OldSorted;                 // set old Sorted
      end;
    end;
  finally
    FreeAndNil(TmpStrings);
  end;
end;

procedure TStringsItem.SetDuplicates(const Value: TDuplicates);
begin
  TStringList(Strings).Duplicates := Value;           // just set Duplicates property
  if UpdateDuplicates then
    RefreshDuplicates;
end;

procedure TStringsItem.SetName(const Value: string);
begin
  FName := Value;
end;

procedure TStringsItem.SetSorted(const Value: Boolean);
begin
  TStringList(Strings).Sorted := Value;
  if UpdateDuplicates then
    RefreshDuplicates;
end;

procedure TStringsItem.SetStrings(const Value: TStrings);
begin
  FStrings.Assign(Value);
  if UpdateDuplicates then
    RefreshDuplicates;
end;

procedure TStringsItem.SetUpdateDuplicates(const Value: Boolean);
begin
  FUpdateDuplicates := Value;
  if UpdateDuplicates then
    RefreshDuplicates;
end;

{ TStringsItems }

function TStringsItems.Add: TStringsItem;
begin
  Result := TStringsItem(inherited Add);
end;

constructor TStringsItems.Create(AStringsContainer: TBcStringsContainer);
begin
  inherited Create(TStringsItem);
  FStringsContainer := AStringsContainer;
end;

function TStringsItems.GetItem(Index: Integer): TStringsItem;
begin
  Result := TStringsItem(inherited GetItem(Index));
end;

function TStringsItems.GetOwner: TPersistent;
begin
  Result := FStringsContainer;
end;

procedure TStringsItems.SetItem(Index: Integer; Value: TStringsItem);
begin
  inherited SetItem(Index, Value);
end;

procedure TStringsItems.SetItemName(Item: TCollectionItem);
begin
//  TStringsItem(Item).Name := Item.ClassName + IntToStr(Item.Index + 1);
  inherited;
end;

procedure TStringsItems.Update(Item: TCollectionItem);
begin
  if (Item <> nil) then ;
end;

{ TBcStringsContainer }

procedure TBcStringsContainer.Clear;
begin
  Items.Clear;
end;

function TBcStringsContainer.ConvertToItemIndex(
  const Index: Integer; var ItemIndex, StringsIndex: Integer): Integer;
var I, CurrentCount, AllCount: Integer;
begin
  { adjust global strings index to local item index and index within the item strings }
  AllCount := 0;
  for I := 0 to Items.Count - 1 do
  begin
    CurrentCount := Items[I].Strings.Count;
    Inc(AllCount, CurrentCount);
    if AllCount > Index then
    begin
      Result := Index - (AllCount - CurrentCount);
      ItemIndex := I;
      StringsIndex := Result;
      Exit;
    end;
  end;
  Result := -1;
  ItemIndex := -1;
  StringsIndex := -1;
end;

constructor TBcStringsContainer.Create(AOwner: TComponent);
begin
  inherited Create(AOwner);
  FItems := TStringsItems.Create(Self);
  FDelimiters := DefaultDelimiters;
  FRaiseErrors := DefaultRaiseErrors;
end;

destructor TBcStringsContainer.Destroy;
begin
  FreeAndNil(FItems);

  inherited Destroy;
end;

function TBcStringsContainer.GetCommaText: string;
var I: Integer;
begin
  Result := '';
  for I := 0 to Count - 1 do
    Result := Result + Items[I].Strings.CommaText;
end;

function TBcStringsContainer.GetCount: Integer;
begin
  Result := Items.Count;
end;

function TBcStringsContainer.GetItemsByName(Name: string): TStringsItem;
var I: Integer;
begin
  Result := nil;
  for I := 0 to Items.Count - 1 do
    if Items[I].Name = Name then
    begin
      Result := Items[I];
      Break;
    end;
end;

function TBcStringsContainer.GetNames(Index: Integer): string;
var I, SI: Integer;
begin
  ConvertToItemIndex(Index, I, SI);
  Result := Items[I].Strings.Names[SI];
end;

function TBcStringsContainer.GetStrings(Index: Integer): string;
var I, SI: Integer;
begin
  ConvertToItemIndex(Index, I, SI);
  Result := Items[I].Strings[SI];
end;

function TBcStringsContainer.GetStringsCount: Integer;
var I: Integer;
begin
  Result := 0;
  for I := 0 to Items.Count - 1 do
    Inc(Result, Items[I].Strings.Count);
end;

function TBcStringsContainer.GetText: string;
var I: Integer;
begin
  Result := '';
  for I := 0 to Count - 1 do
    Result := Result + Items[I].Strings.Text;
end;

function TBcStringsContainer.GetValues(Name: string): string;
var I: Integer;
begin
  Result := '';
  for I := 0 to Items.Count - 1 do
  begin
    Result := Items[I].Strings.Values[Name];
    if Result <> '' then
      Break;
  end;
end;

function TBcStringsContainer.GetValuesByIndex(Index: Integer): string;
var P, I, SI: Integer;
begin
  ConvertToItemIndex(Index, I, SI);
  Result := Items[I].Strings[SI];
  if Length(FDelimiters) > 0 then
    for I := 1 to Length(FDelimiters) do
    begin
      P := AnsiPos(FDelimiters[I], Result);
      if P <> 0 then
      begin
        Delete(Result, 1, P);
        Break;
      end;
    end;
end;

procedure TBcStringsContainer.SaveToFile(const FileName: string);
var
  Strings: TStringList;
  I: Integer;
begin
  try
    Strings := TStringList.Create;
    Strings.Capacity := StringsCount;
    for I := 0 to Items.Count - 1 do
      Strings.AddStrings(Items[I].Strings);
    Strings.SaveToFile(FileName);
  finally
    FreeAndNil(Strings);
  end;
end;

procedure TBcStringsContainer.SetItems(Value: TStringsItems);
begin
  FItems.Assign(Value);
end;

procedure TBcStringsContainer.SetValues(Name: string; const Value: string);
var I, P: Integer;
begin
  for I := 0 to Items.Count - 1 do
  begin
    P := Items[I].Strings.IndexOf(Name);
    if P <> -1 then
    begin
      Items[I].Strings.Values[Name] := Value;
      Exit;
    end;
  end;
  if FRaiseErrors then
    raise EStringListError.CreateFmt('Value %s not found.', [Name]);
end;

end.
